# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.dispatcher.data.packer.base import DataPacker
from typing import List, Any, Dict, Optional, Tuple
import torch
import torch.nn.functional as F
from cosmos_rl.utils.util import retry
from cosmos_rl.policy.config import Config
from cosmos_rl.dispatcher.data.schema import ChatMessage
from transformers import AutoTokenizer, AutoProcessor, AutoConfig
from PIL import Image
import base64
import io

IGNORE_LABEL_ID = -100


def process_vision_info(sample: List[Dict[str, Any]]) -> Tuple[Any, Any]:
    image_inputs = []
    video_inputs = []
    for x in sample:
        if x["role"] == "user":
            for item in x["content"]:
                if item["type"] == "image":
                    image_inputs.append(item["image"])
                if item["type"] == "video":
                    video_inputs.append(item["video"])
    return image_inputs, video_inputs


def encode_image_to_base64(image_inputs: List[str]) -> List[str]:
    new_image_inputs = []
    for image_input in image_inputs:
        img_bytes = base64.b64decode(image_input)
        img_buffer = io.BytesIO(img_bytes)
        image = Image.open(img_buffer)
        new_image_inputs.append(image)
    return new_image_inputs


class Qwen3_VL_DataPacker(DataPacker):
    """
    Data protocol & processing logic for the Qwen3.5 VLMs for SFT and RL training.
    """

    Payload = List[Dict[str, Any]]

    class RLPolicyInput:
        input_ids: List[int]
        logprob_masks: List[int]

        def __init__(self, input_ids: List[int], logprob_masks: List[int]):
            self.input_ids = input_ids
            self.logprob_masks = logprob_masks

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        super().setup(config, tokenizer, *args, **kwargs)
        self.hf_processor = retry(AutoProcessor.from_pretrained)(
            config.policy.model_name_or_path, trust_remote_code=True
        )

        hf_config = retry(AutoConfig.from_pretrained)(
            config.policy.model_name_or_path, trust_remote_code=True
        )

        image_token_id = getattr(hf_config, "image_token_id", None) or getattr(
            hf_config.vision_config, "image_token_id", None
        )
        assert image_token_id is not None, f"Cannot find image token id in {hf_config=}"
        self.image_token_id = image_token_id
        self.image_token = getattr(self.hf_processor, "image_token", None)

        if self.image_token is None:
            self.image_token = self.tokenizer.decode([image_token_id])

        video_token_id = getattr(hf_config, "video_token_id", None) or getattr(
            hf_config.vision_config, "video_token_id", None
        )
        if video_token_id is None:
            self.video_token = None
            self.video_token_id = None
        else:
            self.video_token = self.tokenizer.decode([video_token_id])
            self.video_token_id = video_token_id
        self.vision_ids = [self.image_token_id, self.video_token_id]
        self.hf_config = hf_config

    def get_rollout_input(self, sample: Payload) -> Any:
        """
        This VL data packer only accepts the conversation data format.
        check https://github.com/QwenLM/Qwen2.5-VL?tab=readme-ov-file#using---transformers-to-chat for more details.

        example:
        [
            {
                "role": "user",
                "content": [
                    {
                        "type": "image",
                        "image": "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen-VL/assets/demo.jpeg",
                    },
                    {"type": "text", "text": "Describe this image."},
                ],
            }
        ]

        It is user's responsibility to ensure the conversation format is correct
          and multi-media files involved in conversation are accessible.
        """
        sample = [x.model_dump() if isinstance(x, ChatMessage) else x for x in sample]
        assert all(
            isinstance(x, dict) and "role" in x and "content" in x for x in sample
        ), "All samples should be in conversation format, but got: {}".format(sample)

        if self.image_token is not None:
            for x in sample:
                if x["role"] == "user":
                    contents = x["content"]
                    for idx, content in enumerate(contents):
                        if (
                            content["type"] == "text"
                            and self.image_token in content["text"]
                        ):
                            new_content = content.copy()
                            contents[idx]["text"] = new_content["text"].replace(
                                self.image_token, ""
                            )

        # Here we need to convert the conversation format to the format required by vllm
        prompt = self.hf_processor.apply_chat_template(
            sample, tokenize=False, add_generation_prompt=True
        )
        image_inputs, video_inputs = process_vision_info(sample)
        # TODO: add video support
        if len(video_inputs) > 0:
            return {
                "prompt": prompt,
                "multi_modal_data": {"video": video_inputs},
                "mm_processor_kwargs": {},
            }
        elif len(image_inputs) > 0:
            assert len(image_inputs) == 1, f"{len(image_inputs)=}"
            return {
                "prompt": prompt,
                "multi_modal_data": {"image": image_inputs},
            }
        else:
            return {
                "prompt": prompt,
            }

    def _replace_assistant_content(
        self,
        token_ids: List[int],
        label_ids: List[int],
        pad_token_id: int,
        eos_token_id: int,
        replacement_ids: List[int],
        pad_run_length: int = 10,
    ) -> List[int]:
        """
        Find the first run of exactly `pad_run_length` pad_token_id's in token_ids,
        replace that run with replacement_ids, and return the new list.
        If no such run is found, returns the original list unchanged.
        """
        n = len(token_ids)
        target_run = [pad_token_id] * pad_run_length

        # find the start index of the first matching run
        for i in range(n - pad_run_length + 1):
            if token_ids[i : i + pad_run_length] == target_run:
                # splice in the replacement
                if (
                    len(token_ids) > i + pad_run_length
                    and token_ids[i + pad_run_length] == eos_token_id
                ):
                    label_ids = (
                        label_ids[:i]
                        + replacement_ids
                        + [eos_token_id]
                        + label_ids[i + pad_run_length + 1 :]
                    )
                else:
                    label_ids = (
                        label_ids[:i]
                        + replacement_ids
                        + label_ids[i + pad_run_length :]
                    )
                return (
                    True,
                    token_ids[:i] + replacement_ids + token_ids[i + pad_run_length :],
                    label_ids,
                )
        # no match found
        return False, token_ids, label_ids

    def _get_rope_index(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        image_grid_thw: Optional[torch.LongTensor] = None,
        video_grid_thw: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Different from the original implementation, Qwen3VLMoe use timestamps rather than absolute time position ids."""

        # Since we use timestamps to seperate videos, like <t1> <vision_start> <frame1> <vision_end> <t2> <vision_start> <frame2> <vision_end>, the video_grid_thw should also be split
        if video_grid_thw is not None:
            video_grid_thw = torch.repeat_interleave(
                video_grid_thw, video_grid_thw[:, 0], dim=0
            )
            video_grid_thw[:, 0] = 1

        spatial_merge_size = self.hf_config.vision_config.spatial_merge_size
        image_token_id = self.hf_config.image_token_id
        video_token_id = self.hf_config.video_token_id
        vision_start_token_id = self.hf_config.vision_start_token_id
        mrope_position_deltas = []
        if input_ids is not None and (
            image_grid_thw is not None or video_grid_thw is not None
        ):
            total_input_ids = input_ids
            if attention_mask is None:
                attention_mask = torch.ones_like(total_input_ids)
            position_ids = torch.ones(
                3,
                input_ids.shape[0],
                input_ids.shape[1],
                dtype=input_ids.dtype,
                device=input_ids.device,
            )
            image_index, video_index = 0, 0
            attention_mask = attention_mask.to(total_input_ids.device)
            for i, input_ids in enumerate(total_input_ids):
                input_ids = input_ids[attention_mask[i] == 1]
                image_nums, video_nums = 0, 0
                vision_start_indices = torch.argwhere(
                    input_ids == vision_start_token_id
                ).squeeze(1)
                vision_tokens = input_ids[vision_start_indices + 1]
                image_nums = (vision_tokens == image_token_id).sum()
                video_nums = (vision_tokens == video_token_id).sum()
                input_tokens = input_ids.tolist()
                llm_pos_ids_list: list = []
                st = 0
                remain_images, remain_videos = image_nums, video_nums
                for _ in range(image_nums + video_nums):
                    if image_token_id in input_tokens and remain_images > 0:
                        ed_image = input_tokens.index(image_token_id, st)
                    else:
                        ed_image = len(input_tokens) + 1
                    if video_token_id in input_tokens and remain_videos > 0:
                        ed_video = input_tokens.index(video_token_id, st)
                    else:
                        ed_video = len(input_tokens) + 1
                    if ed_image < ed_video:
                        t, h, w = (
                            image_grid_thw[image_index][0],
                            image_grid_thw[image_index][1],
                            image_grid_thw[image_index][2],
                        )
                        image_index += 1
                        remain_images -= 1
                        ed = ed_image

                    else:
                        t, h, w = (
                            video_grid_thw[video_index][0],
                            video_grid_thw[video_index][1],
                            video_grid_thw[video_index][2],
                        )
                        video_index += 1
                        remain_videos -= 1
                        ed = ed_video
                    llm_grid_t, llm_grid_h, llm_grid_w = (
                        t.item(),
                        h.item() // spatial_merge_size,
                        w.item() // spatial_merge_size,
                    )
                    text_len = ed - st

                    st_idx = (
                        llm_pos_ids_list[-1].max() + 1
                        if len(llm_pos_ids_list) > 0
                        else 0
                    )
                    llm_pos_ids_list.append(
                        torch.arange(text_len).view(1, -1).expand(3, -1) + st_idx
                    )

                    # t_index is always 0 because llm_grid_t is always 1 (we use timestamps to encode the temporal information for videos)
                    t_index = (
                        torch.arange(llm_grid_t)
                        .view(-1, 1)
                        .expand(-1, llm_grid_h * llm_grid_w)
                        .flatten()
                    )
                    h_index = (
                        torch.arange(llm_grid_h)
                        .view(1, -1, 1)
                        .expand(llm_grid_t, -1, llm_grid_w)
                        .flatten()
                    )
                    w_index = (
                        torch.arange(llm_grid_w)
                        .view(1, 1, -1)
                        .expand(llm_grid_t, llm_grid_h, -1)
                        .flatten()
                    )
                    llm_pos_ids_list.append(
                        torch.stack([t_index, h_index, w_index]) + text_len + st_idx
                    )
                    st = ed + llm_grid_t * llm_grid_h * llm_grid_w

                if st < len(input_tokens):
                    st_idx = (
                        llm_pos_ids_list[-1].max() + 1
                        if len(llm_pos_ids_list) > 0
                        else 0
                    )
                    text_len = len(input_tokens) - st
                    llm_pos_ids_list.append(
                        torch.arange(text_len).view(1, -1).expand(3, -1) + st_idx
                    )

                llm_positions = torch.cat(llm_pos_ids_list, dim=1).reshape(3, -1)
                position_ids[..., i, attention_mask[i] == 1] = llm_positions.to(
                    position_ids.device
                )
                mrope_position_deltas.append(
                    llm_positions.max() + 1 - len(total_input_ids[i])
                )
            mrope_position_deltas = torch.tensor(
                mrope_position_deltas, device=input_ids.device
            ).unsqueeze(1)
            return position_ids, mrope_position_deltas
        else:
            if attention_mask is not None:
                position_ids = attention_mask.long().cumsum(-1) - 1
                position_ids.masked_fill_(attention_mask == 0, 1)
                position_ids = (
                    position_ids.unsqueeze(0)
                    .expand(3, -1, -1)
                    .to(attention_mask.device)
                )
                max_position_ids = position_ids.max(0, keepdim=False)[0].max(
                    -1, keepdim=True
                )[0]
                mrope_position_deltas = max_position_ids + 1 - attention_mask.shape[-1]
            else:
                position_ids = (
                    torch.arange(input_ids.shape[1], device=input_ids.device)
                    .view(1, 1, -1)
                    .expand(3, input_ids.shape[0], -1)
                )
                mrope_position_deltas = torch.zeros(
                    [input_ids.shape[0], 1],
                    device=input_ids.device,
                    dtype=input_ids.dtype,
                )

            return position_ids, mrope_position_deltas

    def _process_single_sample(
        self,
        conversation: "Qwen3_VL_DataPacker.Payload",
        add_generation_prompt: bool,
    ) -> Dict[str, Any]:
        try:
            # Replace all the assistant content with consecutive `pad_token` * 10
            pad_token = self.tokenizer.pad_token
            pad_token_id = self.tokenizer.pad_token_id
            eos_token_id = self.tokenizer.eos_token_id
            pad_run_length = 10
            assistant_contents = []
            messages = None
            # SFT
            if "messages" in conversation:
                messages = conversation["messages"]
                for message in messages:
                    if message["role"] == "assistant":
                        content = message["content"]
                        new_content = content.copy()
                        if isinstance(new_content, str):
                            assistant_contents.append(new_content)
                            new_content = pad_token * pad_run_length
                        elif isinstance(new_content, dict):
                            assert (
                                "text" in new_content
                            ), f"text not in content: {content}"
                            assistant_contents.append(new_content["text"])
                            new_content["text"] = pad_token * pad_run_length
                        elif isinstance(content, list):
                            for i, item in enumerate(content):
                                if isinstance(item, dict):
                                    assert (
                                        "text" in item
                                    ), f"text not in content: {item}"
                                    assistant_contents.append(item["text"])
                                    new_content[i]["text"] = pad_token * pad_run_length
                                else:
                                    raise ValueError(
                                        f"Unsupported content type: {type(item)}"
                                    )
                        else:
                            raise ValueError(
                                f"Unsupported content type: {type(content)}"
                            )
                        message["content"] = new_content
            else:
                # RL
                messages = conversation
                if self.image_token is not None:
                    for x in messages:
                        if x["role"] == "user":
                            contents = x["content"]
                            for idx, content in enumerate(contents):
                                if (
                                    content["type"] == "text"
                                    and self.image_token in content["text"]
                                ):
                                    new_content = content.copy()
                                    contents[idx]["text"] = new_content["text"].replace(
                                        self.image_token, ""
                                    )

            text = self.hf_processor.apply_chat_template(
                messages,
                tokenize=False,
                add_generation_prompt=add_generation_prompt,
            )
            if "images" in conversation:
                image_inputs = conversation["images"]
            else:
                image_inputs, video_inputs = process_vision_info(conversation)
                assert all(
                    (isinstance(x, str) for x in image_inputs)
                ), f"{image_inputs=}"
                assert (
                    len(video_inputs) == 0
                ), "Currently video input is not supported for HF VLM"
                image_inputs = encode_image_to_base64(image_inputs)

            kwarg = {
                "return_tensors": "pt",
                "images": image_inputs,
            }
            inputs = self.hf_processor(
                text=[text],
                **kwarg,
            )
            input_ids = inputs["input_ids"][0].tolist()
            label_ids = [IGNORE_LABEL_ID] * len(input_ids)

            for assistant_content in assistant_contents:
                replacement_ids = self.tokenizer.encode(
                    assistant_content, add_special_tokens=False
                )

                replaced, input_ids, label_ids = self._replace_assistant_content(
                    input_ids,
                    label_ids,
                    pad_token_id=pad_token_id,
                    eos_token_id=eos_token_id,
                    replacement_ids=replacement_ids,
                    pad_run_length=pad_run_length,
                )
                if not replaced:
                    raise ValueError("No assistant content to replace")
                if len(input_ids) != len(label_ids):
                    raise ValueError(
                        f"input_ids and label_ids should have the same length, but got {len(input_ids)} and {len(label_ids)}"
                    )
        except Exception as e:
            print(f"Error processing sample: {e}, please fix to ensure SFT works")
            raise e

        result_dict = {
            "input_ids": input_ids,
            "label_ids": label_ids,
        }
        if "pixel_values_videos" in inputs:
            result_dict["pixel_values_videos"] = inputs["pixel_values_videos"]
            if "video_grid_thw" in inputs:
                result_dict["video_grid_thw"] = inputs["video_grid_thw"]
            else:
                result_dict["video_grid_thw"] = None

        if "pixel_values" in inputs:
            result_dict["pixel_values"] = inputs["pixel_values"]
            if "image_grid_thw" in inputs:
                result_dict["image_grid_thw"] = inputs["image_grid_thw"]
            else:
                result_dict["image_grid_thw"] = None

        # position_ids: (3, 1, seq_len)
        # Only for Qwen3VLMOE
        position_ids, _ = self._get_rope_index(
            input_ids=torch.tensor(input_ids).unsqueeze(0).clone(),
            image_grid_thw=torch.tensor(result_dict.get("image_grid_thw"))
            if "image_grid_thw" in result_dict
            else None,
            video_grid_thw=torch.tensor(result_dict.get("video_grid_thw"))
            if "video_grid_thw" in result_dict
            else None,
            attention_mask=None,
        )
        result_dict["position_ids"] = position_ids.clone()

        return result_dict

    def _collate_fn(
        self, processed_samples: List[Dict[str, Any]], computed_max_len: int
    ) -> Dict[str, Any]:
        pixel_values_videos = [x["pixel_values_videos"] for x in processed_samples]
        video_grid_thw = [x["video_grid_thw"] for x in processed_samples]
        pixel_values = [x["pixel_values"] for x in processed_samples]
        image_grid_thw = [x["image_grid_thw"] for x in processed_samples]

        if all([x is not None for x in pixel_values_videos]):
            assert all(
                [x is not None for x in pixel_values_videos]
            ), "pixel_values_videos should not be None"
            max_len = max([x.shape[0] for x in pixel_values_videos])
            for i in range(len(pixel_values_videos)):
                pixel_values_videos[i] = pixel_values_videos[i].unsqueeze(0)
                assert (
                    pixel_values_videos[i].ndim == 3
                ), f"pixel_values_videos[i].ndim: {pixel_values_videos[i].ndim}"
                pixel_values_videos[i] = F.pad(
                    pixel_values_videos[i],
                    (0, 0, 0, max_len - pixel_values_videos[i].shape[1]),
                )
            pixel_values_videos = torch.cat(pixel_values_videos, dim=0)
        else:
            assert all(
                [x is None for x in pixel_values_videos]
            ), "pixel_values_videos should be None"
            pixel_values_videos = None

        if all([x is not None for x in video_grid_thw]):
            video_grid_thw = torch.cat(video_grid_thw, dim=0)
        else:
            assert all(
                [x is None for x in video_grid_thw]
            ), "video_grid_thw should be None"
            video_grid_thw = None

        if all([x is not None for x in pixel_values]):
            pixel_values = torch.cat(pixel_values, dim=0)
        else:
            assert all([x is None for x in pixel_values]), "pixel_values should be None"
            pixel_values = None

        if all([x is not None for x in image_grid_thw]):
            image_grid_thw = torch.cat(image_grid_thw, dim=0)
        else:
            assert all(
                [x is None for x in image_grid_thw]
            ), "image_grid_thw should be None"
            image_grid_thw = None

        # Shape description:
        #
        # pixel_values_[videos/images]: (BATCH_SIZE, N_PATCH, HIDDEN_SIZE)
        # [video/image]_grid_thw: (BATCH_SIZE, 3)
        # second_per_grid_ts: (BATCH_SIZE, 1)
        # pixel_values_[videos/images]_lengths_per_sample: (BATCH_SIZE, 1)
        batch = {}
        if pixel_values_videos is not None:
            batch["pixel_values_videos"] = pixel_values_videos

        if video_grid_thw is not None:
            batch["video_grid_thw"] = video_grid_thw

        if pixel_values is not None:
            batch["pixel_values"] = pixel_values

        if image_grid_thw is not None:
            batch["image_grid_thw"] = image_grid_thw

        # Pad the input_ids, logprob_masks
        batch["input_ids"] = torch.tensor(
            [
                x["input_ids"][:computed_max_len]
                + [self.tokenizer.pad_token_id]
                * (max(0, computed_max_len - len(x["input_ids"])))
                for x in processed_samples
            ],
            dtype=torch.long,
        )
        if "label_ids" in processed_samples[0]:
            batch["label_ids"] = torch.tensor(
                [
                    x["label_ids"][:computed_max_len]
                    + [IGNORE_LABEL_ID]
                    * (max(0, computed_max_len - len(x["label_ids"])))
                    for x in processed_samples
                ],
                dtype=torch.long,
            )
        batch["logprob_masks"] = torch.tensor(
            [
                x["logprob_masks"][:computed_max_len]
                + [0] * (max(0, computed_max_len - len(x["logprob_masks"])))
                for x in processed_samples
            ],
            dtype=torch.bool,
        )

        assert len(batch["input_ids"]) == len(
            batch["logprob_masks"]
        ), "The length of input_ids, logprob_masks should be the same"

        padded_tensors = []
        for sample in processed_samples:
            pad_length = computed_max_len - sample["position_ids"].shape[2]
            padded_tensor = torch.nn.functional.pad(
                sample["position_ids"], (0, pad_length), "constant", 1
            )
            padded_tensors.append(padded_tensor)
        batch["position_ids"] = torch.cat(padded_tensors, dim=1)

        return batch

    def get_policy_input(
        self,
        sample: "Qwen3_VL_DataPacker.Payload",
        rollout_output: Optional[str] = None,
        n_ignore_prefix_tokens: int = 0,
        add_generation_prompt: bool = True,
    ) -> Any:
        # assert all(
        #     isinstance(x, dict) and "role" in x and "content" in x for x in sample
        # ), "All samples should be in conversation format, but got: {}".format(sample)
        x = self._process_single_sample(
            sample,
            add_generation_prompt=add_generation_prompt,
        )

        return_dict = {
            "position_ids": x["position_ids"],
        }
        if "pixel_values_videos" in x:
            return_dict["pixel_values_videos"] = x["pixel_values_videos"]
        else:
            return_dict["pixel_values_videos"] = None

        if "video_grid_thw" in x:
            return_dict["video_grid_thw"] = x["video_grid_thw"]
        else:
            return_dict["video_grid_thw"] = None

        if "pixel_values" in x:
            return_dict["pixel_values"] = x["pixel_values"]
        else:
            return_dict["pixel_values"] = None

        if "image_grid_thw" in x:
            return_dict["image_grid_thw"] = x["image_grid_thw"]
        else:
            return_dict["image_grid_thw"] = None

        # Common fields
        input_ids = x["input_ids"]
        completion_ids = []
        if rollout_output:
            completion_ids = self.tokenizer(rollout_output).input_ids  # Don't pad yet
            # recompute position_ids
            # position_ids: (3, 1, seq_len)
            position_ids, _ = self._get_rope_index(
                input_ids=torch.tensor(input_ids + completion_ids).unsqueeze(0).clone(),
                image_grid_thw=torch.tensor(x.get("image_grid_thw"))
                if "image_grid_thw" in x
                else None,
                video_grid_thw=torch.tensor(x.get("video_grid_thw"))
                if "video_grid_thw" in x
                else None,
                attention_mask=None,
            )
            return_dict["position_ids"] = position_ids.clone()
        return_dict["input_ids"] = input_ids + completion_ids

        return_dict["logprob_masks"] = (
            [0] * (len(input_ids) - 1 + n_ignore_prefix_tokens)
            + [1] * (len(completion_ids) - n_ignore_prefix_tokens)
            + [0]
        )

        return_dict["label_ids"] = x["label_ids"]
        return return_dict

    def policy_compute_max_len(self, processed_samples: List[Dict[str, Any]]) -> int:
        return max([len(x["input_ids"]) for x in processed_samples])

    def policy_collate_fn(
        self, processed_samples: List[Dict[str, Any]], computed_max_len: int
    ) -> Dict[str, Any]:
        for x in processed_samples:
            if "label_ids" in x:
                del x["label_ids"]
        return self._collate_fn(processed_samples, computed_max_len)

    def sft_process_sample(
        self, sample: "Qwen3_VL_DataPacker.Payload"
    ) -> Dict[str, Any]:
        """
        Accepts either raw text or conversation format.
        """
        return self.get_policy_input(sample, add_generation_prompt=False)

    def sft_compute_max_len(self, processed_samples: List[Dict[str, Any]]) -> int:
        """
        Compute the maximum sequence length of the processed samples
        """
        return max([len(x["input_ids"]) for x in processed_samples])

    def sft_collate_fn(
        self,
        processed_samples: List[Dict[str, Any]],
        computed_max_len: int,
        pad_token_id: int,
        ignore_label_id: int,
    ) -> Dict[str, Any]:
        # Reuse the RL collate minibatch function
        model_inputs: Dict[str, Any] = self._collate_fn(
            processed_samples, computed_max_len
        )
        del model_inputs["logprob_masks"]
        # Mask the loss on vision padding tokens
        if self.vision_ids is not None:
            assert isinstance(self.vision_ids, list)
            for vision_id in self.vision_ids:
                if vision_id is not None:
                    model_inputs["label_ids"][
                        model_inputs["label_ids"] == vision_id
                    ] = ignore_label_id

        return model_inputs



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from dataclasses import dataclass
from typing import List, Optional, Tuple, Callable, Union
import torch
import torch.nn as nn
from transformers import AutoConfig
import torch.distributed._symmetric_memory as symm_mem
from cosmos_rl.utils.util import (
    resolve_model_path,
    IdentityLayer,
    clear_weight_name,
    sync_model_vocab,
)
from cosmos_rl.utils.ulysses import (
    slice_inputs_for_ulysses,
)
from cosmos_rl.utils.logging import logger
from safetensors import safe_open
from cosmos_rl.policy.model.internvl.weight_converter import (
    convert_weight_from_hf,
)
from cosmos_rl.policy.model.internvl.weight_mapper import InternVLWeightMapper
from cosmos_rl.policy.kernel.symm_mem_recipes import OnDeviceAllToAllV
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.model.base import ModelRegistry, BaseModel
from functools import cached_property
from cosmos_rl.utils.sequence_packing import pack_sequences_for_inputs
from cosmos_rl.policy.model.gpt import GPTArgs
from cosmos_rl.policy.model.qwen3_moe import (
    Qwen3MoEBlock,
    Qwen3MoeArgs,
    FeedForward,
    RotaryEmbedding as Qwen3MoERotaryEmbedding,
    build_norm as qwen3_moe_build_norm,
)

from cosmos_rl.policy.model.vision_encoder.internvl import (
    InternVL_Encoder_Args,
    InternVisionModel,
)
# from cosmos_rl.dispatcher.data.packer.internvl_data_packer import (
#     InternVL_DataPacker,
# )

InternVL_LM_Args = Union[Qwen3MoeArgs, GPTArgs]


@dataclass
class InternVL_Args:
    lm_args: InternVL_LM_Args
    encoder_args: InternVL_Encoder_Args
    hf_config: AutoConfig = None


class Qwen3MoE(nn.Module):
    def __init__(self, model_args: InternVL_LM_Args):
        super().__init__()
        self.model_args = model_args
        self.vocab_size = model_args.vocab_size
        self.n_layers = model_args.n_layers
        self.rotary_emb = Qwen3MoERotaryEmbedding(model_args)

        self.embed_tokens = nn.Embedding(model_args.vocab_size, model_args.dim)

        self.layers = torch.nn.ModuleDict()
        for layer_id in range(model_args.n_layers):
            self.layers[str(layer_id)] = Qwen3MoEBlock(layer_id, model_args)

        self.norm = qwen3_moe_build_norm(
            model_args.norm_type,
            dim=model_args.dim,
            eps=model_args.norm_eps,
            casting_mode=model_args.hf_config.model_type,
        )

        if not model_args.hf_config.tie_word_embeddings:
            self.tie_embed_tokens = False
            self.lm_head = nn.Linear(
                model_args.dim,
                model_args.vocab_size,
                bias="lm_head" in model_args.biases,
            )
        else:
            self.tie_embed_tokens = True
        self.identity_layer = IdentityLayer()

    def forward(
        self,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        interested_tokens: Optional[torch.BoolTensor] = None,
        **kwargs,  # Additional arguments for compatibility
    ):
        h = self.identity_layer(inputs_embeds)

        position_embeddings = self.rotary_emb(h, position_ids)

        cp_mesh = kwargs.get("cp_mesh", None)

        if "valid_input_len" in kwargs:
            valid_input_len = kwargs["valid_input_len"]
            updated_kwargs = pack_sequences_for_inputs(
                inputs_embeds,
                valid_input_len,
                list(position_embeddings),
                interested_tokens,
                inputs_seq_dim=1,
                inputs_batch_dim=0,
                position_ids_seq_dim=1,
                position_ids_batch_dim=0,
                interested_tokens_seq_dim=1,
                interested_tokens_batch_dim=0,
                padding_mask=kwargs.get("padding_mask", None),
                cp_mesh=cp_mesh,
            )
            position_embeddings = tuple(updated_kwargs.pop("position_ids"))
            interested_tokens = updated_kwargs.pop("interested_tokens")
            h = updated_kwargs.pop("inputs")
            h = self.identity_layer(h)
            kwargs.update(updated_kwargs)
        elif cp_mesh is not None:
            [inputs_embeds, interested_tokens] = slice_inputs_for_ulysses(
                [inputs_embeds, interested_tokens],
                cp_mesh,
                seq_dims=[1, 1],
            )
            position_embeddings = tuple(
                slice_inputs_for_ulysses(
                    list(position_embeddings),
                    cp_mesh,
                    seq_dims=[1] * len(position_embeddings),
                )
            )
            h = self.identity_layer(inputs_embeds)

        for layer in self.layers.values():
            if (
                hasattr(layer, "_gradient_checkpointing_enabled")
                and layer._gradient_checkpointing_enabled
            ):
                h = torch.utils.checkpoint.checkpoint(
                    layer,
                    h,
                    position_embeddings,
                    **kwargs,
                    use_reentrant=False,
                )
            else:
                h = layer(h, position_embeddings=position_embeddings, **kwargs)

        # Add `if` check just in case `pp` is enabled
        if self.norm is not None:
            if interested_tokens is not None:
                assert not isinstance(
                    h, torch.distributed.tensor.DTensor
                ), "interested_tokens must be a local tensor"
                h = h[interested_tokens]
            assert self.lm_head is not None, "lm_head must be provided in last stage"
            h = self.lm_head(self.norm(h))
        return h

    def current_device(self):
        return next(self.parameters()).device

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        for layer in self.layers.values():
            layer.mlp.gate.weight.requires_grad_(False)

        # rotary.inv_freq could get deleted and not re-initialized
        # so we need to delete it manually
        current_device = torch.cuda.current_device()
        self.rotary_emb.to(current_device)
        self.rotary_emb.reset_inv_freq()
        # Basically, max_seq_len * 2 is enough for all-to-all-v communication.
        overflow = 2

        MAX_BATCH_MUL_SEQ_LEN = (
            self.model_args.max_seq_len
            * cosmos_config.train.train_policy.mini_batch
            * self.model_args.hf_config.num_experts_per_tok
        )

        OnDeviceAllToAllV.max_output_len = MAX_BATCH_MUL_SEQ_LEN * overflow
        # Init MoE kernel related buffers
        if FeedForward.token_send_buf is None:
            dtype = self.model_args.hf_config.torch_dtype
            # Input buffer for DP-to-EP shuffle
            FeedForward.token_send_buf = symm_mem.empty(
                MAX_BATCH_MUL_SEQ_LEN,
                self.model_args.dim,  # hidden dim
                dtype=dtype,
                device=self.current_device(),
            )
            FeedForward.token_send_buf.zero_()
            # Input buffer for EP-to-DP shuffle
            FeedForward.token_gather_buf = symm_mem.empty(
                MAX_BATCH_MUL_SEQ_LEN * overflow,
                self.model_args.dim,  # hidden dim
                dtype=dtype,
                device=self.current_device(),
            )
            FeedForward.token_gather_buf.zero_()

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.layers),
            self.model_args.n_heads,
            self.model_args.dim // self.model_args.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )


@ModelRegistry.register(
    InternVLWeightMapper  # , default_data_packer_cls=InternVL_DataPacker
)
class InternVLChatModel(BaseModel):
    def __init__(self, config: InternVL_Args):
        super().__init__(config.hf_config)
        self.config = config
        self.hf_config = config.hf_config
        self.lm_arch = self.hf_config.llm_config.architectures[0]
        self.model = self.get_language_model()
        self.visual = InternVisionModel(config.encoder_args)
        self.vocab_size = config.lm_args.vocab_size

        image_size = (
            self.hf_config.force_image_size or self.hf_config.vision_config.image_size
        )
        patch_size = self.hf_config.vision_config.patch_size
        self.patch_size = patch_size
        self.select_layer = self.hf_config.select_layer
        self.template = self.hf_config.template
        self.ps_version = self.hf_config.ps_version
        self.num_image_token = int(
            (image_size // patch_size) ** 2 * (self.hf_config.downsample_ratio**2)
        )
        self.downsample_ratio = self.hf_config.downsample_ratio
        vit_hidden_size = self.hf_config.vision_config.hidden_size
        llm_hidden_size = self.hf_config.llm_config.hidden_size

        self.mlp1 = nn.Sequential(
            nn.LayerNorm(vit_hidden_size * int(1 / self.downsample_ratio) ** 2),
            nn.Linear(
                vit_hidden_size * int(1 / self.downsample_ratio) ** 2, llm_hidden_size
            ),
            nn.GELU(),
            nn.Linear(llm_hidden_size, llm_hidden_size),
        )

    def get_language_model(self):
        if architecture := self.lm_arch == "Qwen3MoeForCausalLM":
            return Qwen3MoE(self.config.lm_args)
        else:
            raise NotImplementedError(
                f"{architecture} is not implemented for InternVLChatModel."
            )

    def pixel_shuffle(self, x, scale_factor=0.5):
        n, w, h, c = x.size()
        # N, W, H, C --> N, W, H * scale, C // scale
        x = x.view(n, w, int(h * scale_factor), int(c / scale_factor))
        # N, W, H * scale, C // scale --> N, H * scale, W, C // scale
        x = x.permute(0, 2, 1, 3).contiguous()
        # N, H * scale, W, C // scale --> N, H * scale, W * scale, C // (scale ** 2)
        x = x.view(
            n,
            int(h * scale_factor),
            int(w * scale_factor),
            int(c / (scale_factor * scale_factor)),
        )
        if self.ps_version == "v1":
            logger.warning(
                "In ps_version 'v1', the height and width have not been swapped back, "
                "which results in a transposed image."
            )
        else:
            x = x.permute(0, 2, 1, 3).contiguous()
        return x

    def extract_feature(self, pixel_values):
        vit_embeds = self.visual(
            pixel_values=pixel_values, select_layer=self.select_layer
        )
        vit_embeds = vit_embeds[:, 1:, :]

        h = w = int(vit_embeds.shape[1] ** 0.5)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], h, w, -1)
        vit_embeds = self.pixel_shuffle(vit_embeds, scale_factor=self.downsample_ratio)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], -1, vit_embeds.shape[-1])
        vit_embeds = self.mlp1(vit_embeds)
        return vit_embeds

    def _process_vision_embeddings(
        self, input_embeds, input_ids, pixel_values, pad_token_id
    ):
        """Helper function to process vision embeddings (images or videos)"""

        vision_embeds = self.extract_feature(pixel_values)
        batch, seq_len, feature_dim = input_embeds.shape
        input_embeds = input_embeds.reshape(batch * seq_len, feature_dim)
        input_ids = input_ids.reshape(batch * seq_len)
        selected = input_ids == self.image_token_id
        n_image_tokens_in_input_ids = selected.sum()
        n_image_tokens_in_vision_embeds = vision_embeds.reshape(-1, feature_dim).shape[
            0
        ]
        assert (
            n_image_tokens_in_input_ids == n_image_tokens_in_vision_embeds
        ), f"{n_image_tokens_in_input_ids} != {n_image_tokens_in_vision_embeds}"
        input_embeds = input_embeds.clone()
        input_embeds[selected] = (
            vision_embeds.reshape(-1, feature_dim)
            .clone()
            .to(input_embeds.device, input_embeds.dtype)
        )
        input_embeds = input_embeds.reshape(batch, seq_len, feature_dim)

        return input_embeds

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        pixel_values: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        if self.model.embed_tokens is not None:
            assert input_ids.dtype in [
                torch.int32,
                torch.int64,
            ], "input_ids must be of type int32 or int64"
            inputs_embeds = self.model.embed_tokens(input_ids)
            n_image_tokens = (input_ids == self.image_token_id).sum().item()
            # print(f"inputs_embeds: {inputs_embeds.shape}, input_ids: {input_ids.shape}, n_image_tokens: {n_image_tokens}")
            # get vision embeddings as tokens for next phase
            if n_image_tokens > 0:
                inputs_embeds = self._process_vision_embeddings(
                    inputs_embeds,
                    input_ids,
                    pixel_values,
                    self.image_token_id,
                )
        else:
            assert (
                input_ids.is_floating_point()
            ), "input of pipeline stage > 0 must be of floating point type"
            inputs_embeds = input_ids

        # For GRPO, we can pass in the logprob_masks to the model
        # to avoid computing the logits which are not needed for the model
        outputs = self.model(
            inputs_embeds=inputs_embeds,
            position_ids=position_ids,
            interested_tokens=kwargs.pop("interested_tokens", None),
            **kwargs,  # Additional arguments for compatibility
        )
        return outputs

    @property
    def image_token_id(self):
        # TODO: fix this
        # <IMG_CONTEXT> is a special token for image context
        return 151671

    @property
    def video_token_id(self):
        # TODO: fix this
        # <IMG_CONTEXT> is a special token for video context
        return 151671

    @property
    def delay_cp_slice_inputs(self):
        return True

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        if self.lm_arch == "Qwen3MoeForCausalLM":
            seq_dim_idx = 1
            inputs = kwargs["input_ids"]
            position_ids = (
                torch.arange(inputs.size(-1), dtype=torch.long, device=inputs.device)
                .unsqueeze(0)
                .expand_as(inputs)
            )
            return position_ids, inputs, seq_dim_idx
        else:
            raise NotImplementedError(
                f"{self.lm_arch} is not implemented for get_position_ids"
            )

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        self.model.post_to_empty_hook(cosmos_config)

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert pp_size > 1
        is_first = pp_rank == 0
        is_last = pp_rank == pp_size - 1

        # Compute the layers belonging to this stage
        n_layers = len(self.model.layers)
        layers_per_stage = n_layers // pp_size

        if not is_first:
            self.model.embed_tokens = None
            self.visual = None
        if not is_last:
            self.model.lm_head = None
            self.model.norm = None

        local_layers = torch.nn.ModuleDict()
        for i in range(
            pp_rank * layers_per_stage,
            ((pp_rank + 1) * layers_per_stage) if not is_last else n_layers,
        ):
            local_layers[str(i)] = self.model.layers[str(i)]

        # Reset the layers for pipeline splitting
        self.model.layers = local_layers

    @classmethod
    def from_model_args(cls, model_args: InternVL_Args) -> "InternVLChatModel":
        """
        Initialize a GPT model from a GPTArgs object.

        Args:
            model_args (GPTArgs): Model configuration arguments.

        Returns:
            InternVLChatModel: InternVLChatModel model.

        """
        return cls(model_args)

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
        """
        # Load all safetensors from `model_path`
        model_type = self.hf_config.model_type
        lm_type = self.hf_config.llm_config.model_type
        model_path = resolve_model_path(model_name_or_path, revision=revision)
        safetensors_files = [
            f for f in os.listdir(model_path) if f.endswith(".safetensors")
        ]

        # Load LM weights
        lm_state_dict = self.model.state_dict()
        lm_state_dict = {clear_weight_name(k): v for k, v in lm_state_dict.items()}
        # print(f"lm_state_dict: {lm_state_dict.keys()}")
        # Rename dict to remove all `._orig_mod` in keys
        if self.visual is not None:
            visual_state_dict = self.visual.state_dict()
            visual_state_dict = {
                clear_weight_name(k): v for k, v in visual_state_dict.items()
            }
        else:
            visual_state_dict = {}

        multi_modal_projector_state_dict = self.mlp1.state_dict()
        multi_modal_projector_state_dict = {
            clear_weight_name(k): v for k, v in multi_modal_projector_state_dict.items()
        }

        with torch.device(device):
            for f in safetensors_files:
                weights_of_ckpt = {}
                ckpt = safe_open(
                    os.path.join(model_path, f), framework="pt", device=str(device)
                )
                keys = ckpt.keys()
                for name in keys:
                    ckpt_tensor = ckpt.get_tensor(name)
                    weights_of_ckpt[name] = ckpt_tensor

                n_experts = (
                    self.config.lm_args.n_experts if lm_type == "qwen3_moe" else 0
                )

                for name in weights_of_ckpt.keys():
                    tensor = weights_of_ckpt[name]
                    dest_name, shared_weight = convert_weight_from_hf(
                        tensor, name, model_type, lm_type, n_experts, parallel_dims
                    )
                    if dest_name is None:
                        # This is due to the expert parallelism grouping
                        continue
                    # For MoE LM
                    expert_id = None
                    if match := re.search(  # noqa: F841
                        r"layers\.(\d+)\.mlp\.experts\.(\d+)\.(up_proj|gate_proj|down_proj)\.(weight|bias)",
                        dest_name,
                    ):
                        # remove `experts.$ID.` from dest_name
                        expert_id = int(match.group(2))
                        dest_name = dest_name.replace(f"experts.{expert_id}.", "")
                        # Convert expert_id to local_expert_id
                        n_local_experts = (
                            n_experts
                            // parallel_dims.tp
                            // (parallel_dims.dp_shard * parallel_dims.cp)
                        )
                        expert_id = expert_id % n_local_experts

                    if dest_name in lm_state_dict:
                        target_tensor = lm_state_dict[dest_name]
                    elif dest_name in visual_state_dict:
                        target_tensor = visual_state_dict[dest_name]
                    elif dest_name in multi_modal_projector_state_dict:
                        target_tensor = multi_modal_projector_state_dict[dest_name]
                    elif parallel_dims.pp_enabled:
                        # logger.warning(f"Skipping weight: {dest_name} because it's not in the model due to pipeline split")
                        continue
                    else:
                        raise ValueError(f"Unsupported weight: {dest_name}")
                    is_dist_tensor = isinstance(
                        target_tensor, torch.distributed.tensor.DTensor
                    )
                    local_view = (
                        target_tensor.to_local() if is_dist_tensor else target_tensor
                    )
                    # Write to the correct expert of the target tensor
                    if expert_id is not None:
                        local_view = local_view[expert_id]

                    assert (
                        local_view.shape == shared_weight.shape
                    ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name} with original shape {target_tensor.shape}"
                    with torch.no_grad():
                        local_view.data.copy_(shared_weight)

    def separate_model_parts(self) -> List[nn.Module]:
        return [self.model, self.visual, self.mlp1]

    @property
    def parallelize_fn(self) -> Tuple[Callable, nn.Module]:
        from cosmos_rl.policy.model.internvl.parallelize import parallelize

        return parallelize, self

    @staticmethod
    def supported_model_types():
        return ["internvl", "internvl_chat"]

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "InternVLChatModel":
        """
        Initialize a GPT model from a pretrained model.

        Args:
            model_name_or_path (str): Model name or path to the pretrained model.

        Returns:
            InternVLChatModel: InternVLChatModel model.

        """
        if hf_config.model_type not in cls.supported_model_types():
            raise ValueError(f"Unsupported model type: {hf_config.model_type}")

        if max_position_embeddings is None:
            max_position_embeddings = hf_config.max_position_embeddings
        else:
            hf_config.max_position_embeddings = max_position_embeddings

        torch_dtype = hf_config.torch_dtype
        hf_config.llm_config.torch_dtype = torch_dtype
        hf_config.vision_config.torch_dtype = torch_dtype
        vocab_size = sync_model_vocab(model_name_or_path)

        lm_args = None
        if (
            architecture := hf_config.llm_config.architectures[0]
        ) == "Qwen3MoeForCausalLM":
            lm_config = hf_config.llm_config
            # Qwen3MoE does not have any biases
            bias_list = []
            rope_scaling = {}
            if hasattr(lm_config, "rope_scaling"):
                rope_scaling = lm_config.rope_scaling or {}
            rope_type = rope_scaling.get(
                "rope_type", rope_scaling.get("type", "default")
            )
            try:
                head_dim = lm_config.head_dim
            except Exception:
                head_dim = lm_config.hidden_size // lm_config.num_attention_heads
                logger.warning(f"head_dim not found in config, using {head_dim}")

            lm_args = Qwen3MoeArgs(
                dim=lm_config.hidden_size,
                ffn_dim=lm_config.moe_intermediate_size,
                n_layers=lm_config.num_hidden_layers,
                n_experts=lm_config.num_experts,
                n_heads=lm_config.num_attention_heads,
                n_kv_heads=lm_config.num_key_value_heads,
                head_dim=head_dim,
                vocab_size=vocab_size,
                max_seq_len=max_position_embeddings,
                rope_theta=lm_config.rope_theta,
                q_k_norm_enabled=True,
                norm_type="rmsnorm",
                rope_type=rope_type,
                biases=bias_list,
                hf_config=lm_config,
            )
        else:
            raise NotImplementedError(
                f"{architecture} is not implemented for InternVLModel."
            )

        vision_config = hf_config.vision_config
        encoder_args = InternVL_Encoder_Args(
            num_channels=vision_config.num_channels,
            patch_size=vision_config.patch_size,
            image_size=vision_config.image_size,
            qkv_bias=vision_config.qkv_bias,
            hidden_size=vision_config.hidden_size,
            num_attention_heads=vision_config.num_attention_heads,
            intermediate_size=vision_config.intermediate_size,
            qk_normalization=vision_config.qk_normalization,
            num_hidden_layers=vision_config.num_hidden_layers,
            hidden_act=vision_config.hidden_act,
            norm_type=vision_config.norm_type,
            layer_norm_eps=vision_config.layer_norm_eps,
            hf_config=vision_config,
        )
        args = InternVL_Args(
            lm_args=lm_args,
            encoder_args=encoder_args,
            hf_config=hf_config,
        )
        return cls.from_model_args(args)

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        n_params = 0
        n_flops = 0
        if self.visual is not None:
            n_params, n_flops = self.visual._get_nparams_and_flops_fn()(seq_len)
        if self.model is not None:
            lm_n_params, lm_n_flops = self.model._get_nparams_and_flops_fn(seq_len)
            n_params += lm_n_params
            n_flops += lm_n_flops
        return n_params, n_flops

    @classmethod
    def fqn_filter_for_fp8(cls) -> List[str]:
        llm = [
            "lm_head",
        ]
        visual = [
            "visual",
        ]  # Filter Linear in visual out, they will corrupt the FP8 Linear.
        return llm + visual

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        visual_n_heads = self.config.encoder_args.num_attention_heads
        llm_n_heads = self.config.lm_args.n_heads
        cp_compatible = (
            visual_n_heads % (cp_size * tp_size) == 0
            and llm_n_heads % (cp_size * tp_size) == 0
        )
        if not cp_compatible:
            raise ValueError(
                f"Model is not compatible with cp parallelism, model's visual_n_heads={visual_n_heads} or llm_n_heads={llm_n_heads} is not divisible by cp size({cp_size}) * tp_size({tp_size}) = {cp_size * tp_size}"
            )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
For multi-turn messages we provide some helper functions to check and convert to token ids

For the chat template, we follow the chatml format.
To design a powerfull prompt, you can refer https://huggingface.co/docs/transformers/main/en/tasks/prompting
"""

import copy

from transformers import AutoTokenizer
from typing import List, Optional
from cosmos_rl.dispatcher.data.schema import ConversationType, ChatMessage
from cosmos_rl.utils.logging import logger


def get_token_ids_and_loss_mask_from_conversation(
    tokenizer: AutoTokenizer,
    conversation: ConversationType,
    chat_template: Optional[str] = None,
    enable_thinking: bool = False,
    tools: list[dict] = None,
    unmasked_roles: List[str] = ["assistant"],
) -> tuple[List[int], List[int]]:
    """
    Process the multi-turn conversation to token ids and loss mask.

    This function will first replace the unmasked role's message with a placeholder,
    and then replace it with the real message token to generating the final token_ids,
    while also generating an accurate loss_mask.

    Args:
        tokenizer: the tokenizer to use
        conversation: the conversation to process
        enable_thinking: whether to enable thinking
        tools: the tools to use
        unmasked_roles: the roles to unmask, default is only assistant

    Returns:
        concat_tokens_ids: the concatenated tokens
        concat_loss_mask: the concatenated loss mask
    """

    conversation = copy.deepcopy(conversation)
    full_token_ids = tokenizer.apply_chat_template(
        conversation,
        tokenize=True,
        chat_template=chat_template,
        add_generation_prompt=False,
        enable_thinking=enable_thinking,
        tools=tools,
    )
    pad_run_length = 10
    placeholder_tokens = tokenizer.pad_token * pad_run_length
    placeholder_token_ids = [tokenizer.pad_token_id] * pad_run_length

    unmasked_role_contents = []
    # valid the conversation and replace the unmasked roles with placeholder tokens
    for i, msg in enumerate(conversation):
        assert msg.role in [
            "user",
            "assistant",
            "system",
            "tool",
        ], "Unknown role: {}".format(msg.role)
        if msg.role == "system":
            assert i == 0, "System message should be the first message"

        if msg.role in unmasked_roles:
            unmasked_role_contents.append(msg.content)
            msg.content = placeholder_tokens

    if len(unmasked_role_contents) == 0:
        # nothing to replace, return the original token_ids
        return full_token_ids, [0] * len(full_token_ids)

    token_ids_template = tokenizer.apply_chat_template(
        conversation,
        tokenize=True,
        chat_template=chat_template,
        add_generation_prompt=False,
        enable_thinking=enable_thinking,
        tools=tools,
    )

    # Split the token_ids_template by placeholder tokens ids
    token_ids_parts = []
    _idx = 0
    _last_idx = 0
    while _idx < len(token_ids_template) - pad_run_length:
        if token_ids_template[_idx : _idx + pad_run_length] == placeholder_token_ids:
            token_ids_parts.append(token_ids_template[_last_idx:_idx])
            _last_idx = _idx + pad_run_length
            _idx = _last_idx
        else:
            _idx += 1
    token_ids_parts.append(token_ids_template[_last_idx:])

    assert (
        len(token_ids_parts) == len(unmasked_role_contents) + 1
    ), "Number of token ids parts should be the same as the number of unmasked role contents + 1"

    concat_token_ids = []
    concat_loss_mask = []
    for i in range(len(unmasked_role_contents)):
        concat_token_ids.extend(token_ids_parts[i])
        concat_loss_mask.extend([0] * len(token_ids_parts[i]))

        _raw_context_token_ids = tokenizer.encode(
            unmasked_role_contents[i],
            add_special_tokens=False,
        )
        concat_token_ids.extend(_raw_context_token_ids)
        concat_loss_mask.extend([1] * len(_raw_context_token_ids))

    concat_token_ids.extend(token_ids_parts[-1])
    concat_loss_mask.extend([0] * len(token_ids_parts[-1]))

    if full_token_ids != concat_token_ids:
        logger.warning(
            "Full token ids and concat token ids are not the same, use concat token ids instead"
        )
        logger.debug(
            f"Full token ids: {full_token_ids}\nConcat token ids: {concat_token_ids}"
        )
        full_token_ids = concat_token_ids
    return full_token_ids, concat_loss_mask


def add_user_message(messages: ConversationType, content: str) -> ConversationType:
    """
    Add a user message to the conversation.
    """
    messages += [ChatMessage(role="user", content=content)]
    return messages


def add_assistant_message(messages: ConversationType, content: str) -> ConversationType:
    """
    Add an assistant message to the conversation.
    """
    messages += [ChatMessage(role="assistant", content=content)]
    return messages


def add_tool_response_messages(
    messages: ConversationType, tool_response: str
) -> ConversationType:
    """
    Add a tool response message to the conversation.
    """
    messages += [ChatMessage(role="tool", content=tool_response)]
    return messages



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Any
from functools import reduce
from math import gcd


class DimSliceInfo:
    """
    A class to represent the slice information of a tensor along a specific dimension.
    This class contains the offset, total size, dimension name, and length of the slice.
    """

    offset: int
    total_size: int
    dim: str
    length: int = 1

    def __init__(self, offset: int, total_size: int, dim: str = "", length: int = 1):
        """
        Initialize the DimSliceInfo with the given offset, total size, dimension name, and length.
        """
        self.offset = offset
        self.total_size = total_size
        self.dim = dim
        self.length = length

    def __repr__(self):
        # Returning a dictionary representation
        return f"{self.__dict__}"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create a DimSliceInfo object from a dictionary.
        :param data: A dictionary containing the keys 'offset', 'total_size', 'dim', and 'length'.
        :return: A DimSliceInfo object.
        """
        return DimSliceInfo(
            offset=data["offset"],
            total_size=data["total_size"],
            dim=data.get("dim", ""),
            length=data.get("length", 1),
        )

    def simplify(self):
        common = reduce(gcd, [self.offset, self.total_size, self.length])  # noqa: E741
        return DimSliceInfo(
            offset=self.offset // common,
            total_size=self.total_size // common,
            dim=self.dim,
            length=self.length // common,
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from typing import Union
from warnings import warn

import psutil
import torch
from torch import nn
from torch.autograd.graph import saved_tensors_hooks

from cosmos_rl.utils.logging import logger


# We reuse this activation offload manager from torchtune, thanks for this plugin:
# https://github.com/pytorch/torchtune/blob/67ab86b94de9e7ac7dd9850113ebe69e2bbd307c/torchtune/training/_activation_offloading.py#L24


class OffloadActivations(saved_tensors_hooks):
    """Context manager under which activation tensors created in the forward pass will be offloaded.

    Enable the memory efficiency technique of activation offloading, where activations bigger than
    min_offload_size bytes will be offloaded to CPU in the forward and brought back in the backward.
    This is in contrast to maintaining the activation on GPU VRAM throughout the program.

    This manager contains the option of using one additional CUDA stream to handle the communication
    between CUDA and CPU, which is intended to overlap with the default computation stream to improve
    runtime. We designed synchronization with a few heuristics for optimizing the tradeoff between
    runtime vs memory usage.

    Args:
        use_pin_memory (bool): Whether or not the offloaded Tensor will be placed in pinned
            memory on the CPU. Pinned memory allows the Tensor to be moved back onto GPU more quickly
            but is a limited resource. Default: True.

        use_streams (bool): Whether or not to use streams for performance optimization where
            the communications get overlapped with the computation. Requires a torch build
            after torch-2.5.0.]. Default: True.

        max_fwd_stash_size (int): The maximum size of the forward stash, or the maximum number of
            consecutive activations to keep alive during the forward pass. This number must be at
            least 1. Keeping alive more activations will potentially allow more overlap between the
            communication and compute streams at the cost of increasing memory usage. Keeping alive
            fewer activations will conserve memory, but may cause poor overlap between the streams,
            increasing runtime. Default: 5.

        min_offload_size (int): The minimum number of bytes a Tensor must be in order to qualify
            for offloading. If the tensor is too small, we do not want to waste bandwidth and resources
            moving it to CPU and back. Default: 1024 bytes.

    Raises:
        ValueError: if max_fwd_stash_size is not at least 1.

    Example:
        >>> with OffloadActivations():
        >>>     logits = model(inputs)
        >>> loss = ...
        >>> loss.backward()
    """

    def __init__(
        self,
        use_pin_memory: bool = True,
        use_streams: bool = True,
        max_fwd_stash_size: int = 5,
        min_offload_size: int = 1024,
    ) -> None:
        self.use_streams: bool = use_streams

        self.min_tensor_size_bytes = (
            min_offload_size  # we don't want to bother with small tensors
        )
        self.tracker = {}  # tensor_id => (new_tensor, if_modified)  ---> track what saved/offloaded tensors are where
        self.tensor_id: int = 0
        self.is_first_forward_call = True
        self.is_first_backward_call = True
        self.is_first_forward_pass = True

        # managing cpu memory
        self.use_pin_memory: bool = use_pin_memory
        self.virtual_memory_safe_pct = (
            60  # we should not exceed this percentage of memory
        )
        # comp stream
        self.s0 = torch.cuda.current_stream()

        # for streaming
        if self.use_streams:
            self.s1 = torch.cuda.Stream()  # comms stream
            self.fwd_stash = {}  # tensor_id => (activation, ev1)
            if max_fwd_stash_size < 1:
                raise ValueError(
                    f"max_fwd_stash_size should be at least 1 but is {max_fwd_stash_size}"
                )
            self.max_fwd_stash_size = max_fwd_stash_size
            self.bwd_tensor_stash = {}  # tensor_id => activation
            self.bwd_ev_stash = {}  # tensor_id => ev0
            self.curr_graph_id = None
            self.curr_autograd_node = None

        # -------- platform util functions -------- #
        def verify_sufficient_virtual_memory():
            curr_pct = get_cpu_ram_pct()
            if curr_pct > self.virtual_memory_safe_pct:
                warn(
                    f"***** WARNING: {curr_pct=}% > {self.virtual_memory_safe_pct=}% of virtual memory used"
                )

        def get_cpu_ram_pct() -> float:
            # get the percentage of memory used by the system
            return psutil.virtual_memory().percent

        def get_tensor_id() -> int:
            # create a unique id for each tensor we are managing
            self.tensor_id += 1
            return self.tensor_id

        def get_num_bytes_tensor(x: torch.Tensor) -> int:
            # get the number of bytes in a tensor, for memory management purposes
            return (
                x.element_size() * x.nelement()
            )  # x.element_size() * x._base_storage().nbytes()

        # -------- core pack / unpack work -------- #
        def pack_tensor(activation: torch.Tensor) -> int:
            # activations are passed in during forward pass - from here we take over and return a unique id
            if self.is_first_forward_call:
                assert (
                    len(self.tracker) == 0
                ), "backward pass should have cleared tracker of all tensors"

                # set training phase trackers
                self.is_first_forward_call = False
                self.is_first_backward_call = True

            # query for basic tensor info
            num_bytes = get_num_bytes_tensor(activation)
            tensor_id = get_tensor_id()

            # only offload hefty bois if they're activations on CUDA (our heuristic
            # for that is to check if they're not params or buffers)!
            if (
                activation.device.type == torch.accelerator.current_accelerator().type
                and num_bytes >= self.min_tensor_size_bytes
                and (
                    not isinstance(activation, torch.nn.Parameter)
                    and not (
                        hasattr(torch.nn, "Buffer")
                        and isinstance(activation, torch.nn.Buffer)
                    )
                )
            ):
                if self.use_streams:
                    # First, sync back and dereference previously offloaded tensors
                    # as the offloading should be done sufficiently long ago.
                    for id in [k for k in self.fwd_stash.keys()]:
                        if id <= tensor_id - self.max_fwd_stash_size:
                            _, ev = self.fwd_stash[id]
                            self.s0.wait_event(ev)
                            del self.fwd_stash[id]
                        else:
                            break

                    # Sync in, offload, and add an event to sync back later
                    self.s1.wait_stream(self.s0)

                stream = self.s1 if self.use_streams else self.s0
                with stream:
                    cpu_tensor = torch.empty_like(
                        activation, pin_memory=self.use_pin_memory, device="cpu"
                    )
                    cpu_tensor.copy_(activation, non_blocking=True)
                    self.tracker[tensor_id] = (
                        cpu_tensor,
                        True,
                    )  # True = (in future) modified

                if self.use_streams:
                    event = self.s1.record_event()

                    # Stash to keep activation alive untill s1 is done
                    self.fwd_stash[tensor_id] = (activation, event)
            else:
                self.tracker[tensor_id] = (
                    activation,
                    False,
                )  # False = not modified, tensor is as is

            return tensor_id

        def unpack_tensor_single_stream(unpack_tensor_id: int) -> torch.Tensor:
            # backward pass - we are called with the tensor_id, which
            # we will use to retrieve the saved/offloaded tensor
            if self.is_first_backward_call:
                if self.is_first_forward_pass:
                    self.is_first_forward_pass = False
                    if self.use_pin_memory:
                        verify_sufficient_virtual_memory()

                self.is_first_backward_call = False
                self.is_first_forward_call = True

            assert (
                unpack_tensor_id in self.tracker
            ), f"untracked tensor with id {unpack_tensor_id}"

            maybe_gpu_tensor, modified = self.tracker[unpack_tensor_id]
            if modified:
                gpu_tensor = maybe_gpu_tensor.to(
                    torch.accelerator.current_accelerator(), non_blocking=True
                )
                maybe_gpu_tensor = gpu_tensor

            # clear tensor from tracking
            del self.tracker[unpack_tensor_id]
            return maybe_gpu_tensor

        def unpack_tensor_with_streams(unpack_tensor_id: int) -> torch.Tensor:
            # backward pass - we are called with the tensor_id, which
            # we will use to retrieve the saved/offloaded tensor
            if self.is_first_backward_call:
                self.curr_graph_id = torch._C._current_graph_task_id()

                def wait_and_del_remaining_references() -> None:
                    for id in [k for k in self.bwd_tensor_stash.keys()]:
                        event = self.bwd_ev_stash[id]
                        self.s1.wait_event(event)
                        del self.bwd_tensor_stash[id]

                # Register a callback to the end of autograd to clean everything up
                torch.autograd.variable.Variable._execution_engine.queue_callback(
                    wait_and_del_remaining_references
                )

                if self.is_first_forward_pass:
                    self.is_first_forward_pass = False
                    if self.use_pin_memory:
                        verify_sufficient_virtual_memory()

                self.is_first_backward_call = False
                self.is_first_forward_call = True

            assert (
                unpack_tensor_id in self.tracker
            ), f"untracked tensor with id {unpack_tensor_id}"

            maybe_gpu_tensor, modified = self.tracker[unpack_tensor_id]
            if modified:
                # Get data on the current autograd node
                graph_id = torch._C._current_graph_task_id()
                node = torch._C._current_autograd_node()
                prev_node_ids = []

                # If we're on a new node, mark prev node's tensors to be freed later
                if graph_id == self.curr_graph_id and self.curr_autograd_node != node:
                    self.curr_autograd_node = node
                    prev_node_ids = [id for id in self.bwd_tensor_stash.keys()]

                brought_back_from_cpu = True
                if unpack_tensor_id in self.fwd_stash:
                    # this means that the tensor could be just in the process of being offloaded
                    maybe_gpu_tensor = self.fwd_stash[unpack_tensor_id][0]
                    brought_back_from_cpu = False
                else:
                    # Kick off the process to bring tensors back
                    with self.s1:
                        gpu_tensor = maybe_gpu_tensor.to(
                            torch.accelerator.current_accelerator(), non_blocking=True
                        )
                        maybe_gpu_tensor = gpu_tensor

                    # Tell comp stream to wait for the info to be loaded before executing
                    self.s0.wait_stream(self.s1)

                    # Stash the tensor to keep memory alive until compute stream is complete
                    self.bwd_tensor_stash[unpack_tensor_id] = maybe_gpu_tensor

                    # Note: [Track views of the unpacked]
                    # Why do we get the use count of the unpacked tensor here? We want an
                    # initial count to compare to later, during the post-hook of the
                    # backward node, when we need to decide whether we're allowed to free
                    # the tensor yet. In what obscure cases must we delay freeing the
                    # tensor (and thus call record_stream)?
                    # 1. Any of the outputs of the backward node is a view of the unpacked
                    #    tensor.
                    # 2. In the case that this unpacked tensor will be used in a
                    #    checkpointed region, if one of the recomputed saved tensors ends
                    #    up as a view of the unpacked tensor.
                    # 3. The user abuses the system somehow and manually relies on the
                    #    unpacked tensor to exist after the backward node has executed.
                    storage_refcount = torch._C._storage_Use_Count(
                        maybe_gpu_tensor.untyped_storage()._cdata
                    )

                def hook(outputs, inputs):
                    # create events for the current node inputs/outputs if they were streamed in
                    if brought_back_from_cpu:
                        # See Note: [Track views of the unpacked]
                        # IF any of the outputs is a view of the tensor, OR if a view of
                        # the tensor has been saved as a part of checkpoint's recompute
                        # process, OR the user has abusedly incurred a reference on the
                        # unpacked tensor, THEN the tensor might be used later and we
                        # cannot presume to delete it after only the current node is
                        # done! So we use our frenemy, record_stream, to ensure the
                        # Tensor stays unmessed with until it's done getting used in the
                        # compute stream (s0 here). Note that the con here is we introduce
                        # non-deterministic (thus higher) memory usage, but this case
                        # should not happen often.
                        unpacked_tensor = self.bwd_tensor_stash[unpack_tensor_id]
                        if (
                            torch._C._storage_Use_Count(
                                unpacked_tensor.untyped_storage()._cdata
                            )
                            > storage_refcount
                        ):
                            unpacked_tensor.record_stream(self.s0)
                            del self.bwd_tensor_stash[unpack_tensor_id]
                        else:
                            event = self.s0.record_event()
                            self.bwd_ev_stash[unpack_tensor_id] = event

                    # if there are still things in the fwd_stash, get rid of them as we're in bwd now
                    for id in [k for k in self.fwd_stash.keys()]:
                        _, ev = self.fwd_stash[id]
                        self.s0.wait_event(ev)
                        del self.fwd_stash[id]

                    # wait on prev node's events and del those
                    for id in prev_node_ids:
                        event = self.bwd_ev_stash[id]
                        self.s1.wait_event(event)
                        del self.bwd_tensor_stash[id]

                    return outputs

                node.register_hook(hook)

            # clear tensor from tracking
            del self.tracker[unpack_tensor_id]
            return maybe_gpu_tensor

        unpack_tensor = (
            unpack_tensor_with_streams
            if self.use_streams
            else unpack_tensor_single_stream
        )
        super().__init__(pack_tensor, unpack_tensor)


class NoOpManager(saved_tensors_hooks):
    """
    A saved_tensors_hook manager used to disable any other saved_tensors_hook manager
    applied before. This relies on the behavior that only the most recently registered
    saved_tensors_hook will run.

    One example usage is to opt a local region of code out of activations offloading,
    which is usually applied globally to best track state.
    """

    def __init__(self) -> None:
        def noop(tensor):
            return tensor

        super().__init__(noop, noop)


def get_act_offloading_ctx_manager(
    model: nn.Module, enable_activation_offloading: bool, use_streams: bool = True
) -> Union[OffloadActivations, contextlib.nullcontext]:
    """Returns the activation offloading context manager for the model, which will be
    a null context if enable_activation_offloading is False.

    If activation offloading is enabled, we return the OffloadActivations context manager.
    If activation offloading is disabled, we return a NoOpManager context manager.

    Args:
        model (nn.Module): the model to wrap with the activation offloading context manager.
        enable_activation_offloading (bool): whether or not to enable activation offloading
            for the model.
        use_streams (bool): whether or not to enable streams for overlapping communication.

    Returns:
        contextlib.ContextDecorator: the activation offloading context manager for the model.

    Raises:
        NotImplementedError: If the model is a multimodal model and activation offloading is enabled.
    """
    if enable_activation_offloading:
        activations_handling_ctx = OffloadActivations(use_streams=use_streams)

        # Below is our hack to disable offloading the last output Linear in every
        # step, as the cost for offloading the activation and then soon after bringing
        # it back is expensive. Moreover, due to heuristics in our streaming API,
        # we actually use more memory if we offload it as it interferes with chunkedCE.
        lm_head_detected = False
        noop_ctx = NoOpManager()
        if hasattr(model, "lm_head"):
            if isinstance(model.lm_head, nn.Module):
                model.lm_head.register_forward_pre_hook(
                    lambda *args: noop_ctx.__enter__()
                )
                model.lm_head.register_forward_hook(
                    lambda *args: noop_ctx.__exit__(), always_call=True
                )
                lm_head_detected = True
        # FIXME: (lms) Add the check for the VLM.
        if not lm_head_detected:
            logger.warning(
                "During activation offloading, no lm_head was detected. "
                "If your model has an lm head, it will be offloaded. "
                "This usually greatly slows training, given the large vocabulary size. "
                "To change this behavior, set your lm head as model.lm_head and make it "
                "an nn.Module."
            )

    else:
        activations_handling_ctx = contextlib.nullcontext()

    return activations_handling_ctx



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from typing import Dict, Any
from vllm.model_executor.layers.quantization.mxfp4 import Mxfp4MoEMethod
from vllm.model_executor.layers.fused_moe import FusedMoE
from cosmos_rl.utils.dim_slice_info import DimSliceInfo
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.arithmetic import cdiv


def genereate_dim_rank_info(
    module: FusedMoE,
    param_name: str,
    param: torch.Tensor,
    hf_config: Any,
    parallel_dims: ParallelDims,
) -> Dict[int, DimSliceInfo]:
    """
    Generate the dim_rank_info for the parameter. Main logic is from vLLM sharding strategy for GPT-OSS MXFP4.
    """
    dim_rank_info = {}
    mxfp4_block = 32
    intermediate_size = hf_config.intermediate_size
    tp_rank, tp_size = parallel_dims.tp_coord
    quant_method = getattr(module, "quant_method", None)
    assert quant_method is not None, "Must have quant_method for FusedMoE"
    assert isinstance(
        quant_method, Mxfp4MoEMethod
    ), "Only mxfp4 quant method is supported for now."
    # get the rounded shape of weight
    # 768
    intermediate_size_per_partition_after_pad = quant_method.intermediate_size
    # hidden_size_of_quant_method = quant_method.hidden_size  # This may also padded.

    weight_intermediate_size = intermediate_size_per_partition_after_pad
    scale_factor = 1
    tp_dim = -1
    if "w13_weight" in param_name:
        # shape: [num_experts, intermediate_size_per_partition_after_pad * 2, hidden_size // 2]
        tp_dim = 1
        scale_factor = 2
        weight_intermediate_size *= 2
    elif "w2_weight" in param_name:
        # shape: [num_experts, hidden_size, intermediate_size_per_partition_after_pad // 2]
        tp_dim = 2
    elif "w13_bias" in param_name:
        # shape: [num_experts, intermediate_size_per_partition_after_pad * 2]
        weight_intermediate_size *= 2
        scale_factor = 2
        tp_dim = 1
    elif "w2_bias" in param_name:
        # for w2_bias, each rank will take full shards.
        # But only rank 0 will load the full weight, other ranks will set w2_bias to zeros.
        # shape: [num_experts, hidden_size]
        pass
    else:
        pass

    # determin the lenght of dimsliceinfo
    intermediate_size_block = intermediate_size // mxfp4_block
    per_rank_intermediate_size_block = cdiv(intermediate_size_block, tp_size)
    per_rank_intermediate_size = per_rank_intermediate_size_block * mxfp4_block

    tp_rank_start = tp_rank * per_rank_intermediate_size
    tp_rank_end = min((tp_rank + 1) * per_rank_intermediate_size, intermediate_size)

    length = tp_rank_end - tp_rank_start
    weight_shape = param.shape
    for idx, g_size in enumerate(weight_shape):
        if idx == tp_dim:
            dim_rank_info[idx] = DimSliceInfo(
                offset=scale_factor * tp_rank_start,
                total_size=scale_factor * intermediate_size,
                length=scale_factor * length,
            ).__dict__
        else:
            dim_rank_info[idx] = DimSliceInfo(
                offset=0,
                total_size=g_size,
                length=g_size,
            ).__dict__

    return dim_rank_info, tp_dim



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import re
from typing import List, Tuple, Dict, Any
from cosmos_rl.policy.model.base import WeightMapper
from cosmos_rl.utils import util
from transformers import AutoConfig
from functools import cached_property


class HFModelWeightMapper(WeightMapper):
    def __init__(self, hf_config: AutoConfig):
        super().__init__(hf_config)
        self.kv_head_ratio = 1
        self.head_dim = 1
        if getattr(self.config, "num_key_value_heads", None) is not None:
            self.kv_head_ratio = (
                self.config.num_attention_heads // self.config.num_key_value_heads
            )
            self.head_dim = self.config.hidden_size // self.config.num_attention_heads
        elif getattr(self.config, "text_config", None) is not None:
            # VLM models like Gemma3-12b-it has num_attention_heads in text_config
            text_config = self.config.text_config
            self.kv_head_ratio = (
                text_config.num_attention_heads // text_config.num_key_value_heads
            )
            self.head_dim = text_config.hidden_size // text_config.num_attention_heads
        elif getattr(self.config, "llm_config", None) is not None:
            # VLM models like InternVL could has num_attention_heads in llm_config
            text_config = self.config.llm_config
            self.kv_head_ratio = (
                text_config.num_attention_heads // text_config.num_key_value_heads
            )
            self.head_dim = text_config.hidden_size // text_config.num_attention_heads
        else:
            raise ValueError(
                f"Can not determine kv_head_ratio and head_dim from config: {self.config}"
            )
        self.is_vlm = getattr(self.config, "vision_config", None) is not None
        self.reverse_hf_conversion_mapping = None

    def _rollout_vllm_name_to_hf(self, rollout_weight_name: str) -> str:
        # Happen to be the same as policy name mapping.
        model_type = self.config.model_type
        if model_type == "gpt_oss":
            # Some special cases for GPT-OSS.
            gpt_oss_rename_mapping = {
                # Please do not change the order of the keys.
                "attn": "self_attn",
                "embedding": "embed_tokens",
            }
            for key, value in gpt_oss_rename_mapping.items():
                if key in rollout_weight_name:
                    return rollout_weight_name.replace(key, value)
            # gate_up_proj
            if "w13_weight" in rollout_weight_name:
                return rollout_weight_name.replace("w13_weight", "gate_up_proj")
            elif "w2_weight" in rollout_weight_name:
                return rollout_weight_name.replace("w2_weight", "down_proj")
            elif "w13_bias" in rollout_weight_name:
                return rollout_weight_name.replace("w13_bias", "gate_up_proj_bias")
            elif "w2_bias" in rollout_weight_name:
                return rollout_weight_name.replace("w2_bias", "down_proj_bias")
            else:
                pass
        elif model_type == "qwen3_vl":
            # Special case for Qwen3-VL
            if rollout_weight_name.startswith("language_model.model."):
                rollout_weight_name = rollout_weight_name.replace(
                    "language_model.model.", "language_model."
                )
            if (
                not rollout_weight_name.startswith("model.")
                and "lm_head" not in rollout_weight_name
            ):
                rollout_weight_name = "model." + rollout_weight_name
            if rollout_weight_name.startswith("language_model.lm_head."):
                # lm_head exists at language_model level, remove the language_model prefix
                rollout_weight_name = rollout_weight_name.replace(
                    "language_model.lm_head.", "lm_head."
                )

            return rollout_weight_name

        return self.policy_map_local_key_to_hf_key(rollout_weight_name)

    def _rollout_split_qkv_weight(self, name, weight: torch.Tensor):
        if "visual" in name or "vision_tower" in name:
            # split qkv weight for visual
            # weight has shape [3 * head_dim, hidden_dim]
            # kv head ratio is 1, so we can split it into q, k, v
            assert (
                weight.shape[0] % 3 == 0
            ), "Weight shape is not compatible for splitting."
            unit_dim = weight.shape[0] // 3  # for both weight and bias
            q_weight = weight[:unit_dim]
            k_weight = weight[unit_dim : unit_dim * 2]
            v_weight = weight[unit_dim * 2 :]
            return q_weight, k_weight, v_weight
        # weight has shape [q_num_heads * head_dim + k_num_heads * head_dim + v_num_heads * head_dim, hidden_dim]
        # bias has shape [(q_num_heads + k_num_heads + v_num_heads) * head_dim]
        shares = self.kv_head_ratio + 2
        dim_0 = weight.shape[0]  # for both weight and bias
        unit_dim = dim_0 // shares

        q_weight = weight[: unit_dim * self.kv_head_ratio]
        k_weight = weight[
            unit_dim * self.kv_head_ratio : unit_dim * (self.kv_head_ratio + 1)
        ]
        v_weight = weight[unit_dim * (self.kv_head_ratio + 1) :]
        return q_weight, k_weight, v_weight

    def _split_gate_proj_weight(self, name, weight: torch.Tensor):
        # weight has shape [2 * x, hidden_dim]
        dim_0 = weight.shape[0]
        gate_proj_weight = weight[: dim_0 // 2]
        up_proj_weight = weight[dim_0 // 2 :]
        return gate_proj_weight, up_proj_weight

    def rollout_prepare_recv(
        self, vllm_model: Any
    ) -> Tuple[Dict[str, torch.Tensor], List[Tuple[str, torch.Size]]]:
        models_do_not_split_gate_up_proj = ["gpt_oss"]
        recv_key_n_shape_list = []
        vllm_weight_inplace_view_map = {}
        # For some models like Qwen3-VL, vLLM just soft link lm_head with embed_tokens
        # Like: `self.lm_head = self.model.embed_tokens`
        # If we don't set `remove_duplicate` to False, the `lm_head` will not be included in the named_parameters.
        for param_name, param in vllm_model.named_parameters(remove_duplicate=False):
            group_keys = []
            compatible_key = self._rollout_vllm_name_to_hf(param_name)
            if ("qkv_proj" in compatible_key) or (
                "qkv" in compatible_key and not self.is_vlm
            ):
                # must be inplace slicing.
                # split qkv weight
                rule = "qkv_proj" if "qkv_proj" in compatible_key else "qkv"
                q_weight, k_weight, v_weight = self._rollout_split_qkv_weight(
                    compatible_key, param
                )
                q_proj_weight_key = compatible_key.replace(rule, "q_proj")
                k_proj_weight_key = compatible_key.replace(rule, "k_proj")
                v_proj_weight_key = compatible_key.replace(rule, "v_proj")

                vllm_weight_inplace_view_map[q_proj_weight_key] = q_weight
                group_keys.append((q_proj_weight_key, q_weight.ndim))
                vllm_weight_inplace_view_map[k_proj_weight_key] = k_weight
                group_keys.append((k_proj_weight_key, k_weight.ndim))
                vllm_weight_inplace_view_map[v_proj_weight_key] = v_weight
                group_keys.append((v_proj_weight_key, v_weight.ndim))
            elif (
                "gate_up_proj" in compatible_key
                and self.config.model_type not in models_do_not_split_gate_up_proj
            ):
                # split gate and up proj
                gate_proj_weight, up_proj_weight = self._split_gate_proj_weight(
                    compatible_key, param
                )
                gate_proj_weight_key = compatible_key.replace(
                    "gate_up_proj", "gate_proj"
                )
                vllm_weight_inplace_view_map[gate_proj_weight_key] = gate_proj_weight
                group_keys.append((gate_proj_weight_key, gate_proj_weight.ndim))

                up_proj_weight_key = compatible_key.replace("gate_up_proj", "up_proj")
                vllm_weight_inplace_view_map[up_proj_weight_key] = up_proj_weight
                group_keys.append((up_proj_weight_key, up_proj_weight.ndim))
            elif "qkv" in compatible_key:
                q_weight, k_weight, v_weight = self._rollout_split_qkv_weight(
                    compatible_key, param
                )
                q_visual_proj_weight_key = compatible_key.replace("qkv", "q")
                k_visual_proj_weight_key = compatible_key.replace("qkv", "k")
                v_visual_proj_weight_key = compatible_key.replace("qkv", "v")
                vllm_weight_inplace_view_map[q_visual_proj_weight_key] = q_weight
                group_keys.append((q_visual_proj_weight_key, q_weight.ndim))
                vllm_weight_inplace_view_map[k_visual_proj_weight_key] = k_weight
                group_keys.append((k_visual_proj_weight_key, k_weight.ndim))
                vllm_weight_inplace_view_map[v_visual_proj_weight_key] = v_weight
                group_keys.append((v_visual_proj_weight_key, v_weight.ndim))
            else:
                vllm_weight_inplace_view_map[compatible_key] = param
                group_keys.append((compatible_key, param.ndim))

            recv_key_n_shape_list.append(group_keys)
        return vllm_weight_inplace_view_map, recv_key_n_shape_list

    def policy_map_local_key_to_hf_key(self, name: str) -> str:
        name = util.clear_weight_name(name)
        if self.is_vlm:
            # Policy model is HFModel, so we need to reverse the hf checkpoint conversion mapping
            if self.reverse_hf_conversion_mapping:
                for pattern, replacement in self.reverse_hf_conversion_mapping.items():
                    if re.match(pattern, name):
                        name = re.sub(pattern, replacement, name)
                        break
            else:
                # Rollout model is vllm model, so we don't need to reverse the hf checkpoint conversion mapping
                pass
        else:
            if not name == "lm_head.weight":
                if not name.startswith("model."):
                    name = "model." + name
        return name

    def name_to_model_part_index(self, dest_name: str) -> int:
        if dest_name in ["lm_head.weight", "lm_head.bias"]:
            return 0
        elif dest_name.startswith("visual.") or dest_name.startswith("vision_tower."):
            return 1
        elif dest_name.startswith("model.") or dest_name.startswith("language_model."):
            return 0
        else:
            raise ValueError(f"Unsupported weight: {dest_name}")

    def policy_decompose_param_1_to_n_for_sync(self, name):
        """
        Map a parameter of the policy model to set of transformed parameters that need to be synchronized.
        This method returns a list containing tuples of the new parameter name and the corresponding new tensor transformed from the original tensor of the given name.
        Each tuple element includes a transformed tensor and its corresponding slice strategy to derive from the original tensor.
        """
        # Addedd for models with qkv_proj and gate_up_proj like phi4
        if match := re.search(  # noqa: F841
            r"model\.layers\.(\d+)\.self_attn\.qkv_proj\.(weight|bias)",
            name,
        ):
            total_size = self.kv_head_ratio + 2
            split_strategy = []
            # The first part of the split:
            # the dictionary means at dimension 0, extract the part of offset 0 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv_proj", "q_proj"),
                    {
                        0: {
                            "offset": 0,
                            "total_size": total_size,
                            "length": self.kv_head_ratio,
                        }
                    },
                )
            )
            # The second part of the split:
            # the dictionary means at dimension 0, extract the part of offset 1 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv_proj", "k_proj"),
                    {
                        0: {
                            "offset": self.kv_head_ratio,
                            "total_size": total_size,
                            "length": 1,
                        }
                    },
                )
            )
            # The third part of the split:
            # the dictionary means at dimension 0, extract the part of offset 2 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv_proj", "v_proj"),
                    {
                        0: {
                            "offset": self.kv_head_ratio + 1,
                            "total_size": total_size,
                            "length": 1,
                        }
                    },
                )
            )
            return split_strategy
        elif match := re.search(  # noqa: F841
            r"(visual|vision_tower)\.blocks\.(\d+)\.attn\.qkv\.(weight|bias)",
            name,
        ):
            split_strategy = []
            # The first part of the split:
            # the dictionary means at dimension 0, extract the part of offset 0 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "q"),
                    {0: {"offset": 0, "total_size": 3, "length": 1}},
                )
            )
            # The second part of the split:
            # the dictionary means at dimension 0, extract the part of offset 1 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "k"),
                    {0: {"offset": 1, "total_size": 3, "length": 1}},
                )
            )
            # The third part of the split:
            # the dictionary means at dimension 0, extract the part of offset 2 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "v"),
                    {0: {"offset": 2, "total_size": 3, "length": 1}},
                )
            )
            return split_strategy
        elif match := re.search(  # noqa: F841
            r"model\.layers\.(\d+)\.mlp\.gate_up_proj\.(weight|bias)",
            name,
        ):
            split_strategy = []
            split_strategy.append(
                (
                    name.replace("gate_up_proj", "gate_proj"),
                    {0: {"offset": 0, "total_size": 2, "length": 1}},
                )
            )
            split_strategy.append(
                (
                    name.replace("gate_up_proj", "up_proj"),
                    {0: {"offset": 1, "total_size": 2, "length": 1}},
                )
            )
            return split_strategy
        return []

    @cached_property
    def packed_modules_mapping(self):
        mapping_dict = {
            "qkv": [
                "q",
                "k",
                "v",
            ],
            "gate_up_proj": [
                "gate_proj",
                "up_proj",
            ],
            "qkv_proj": [
                "q_proj",
                "k_proj",
                "v_proj",
            ],
        }
        if self.config.model_type == "gpt_oss":
            mapping_dict["qkv"] = ["q_proj", "k_proj", "v_proj"]
        return mapping_dict

    def get_unsplited_weight_name(self, weight_key: str) -> str:
        for key in ["q_proj", "k_proj", "v_proj"]:
            if key in weight_key:
                if "gpt_oss" in self.config.model_type:
                    return weight_key.replace(key, "qkv")
                else:
                    return weight_key.replace(key, "qkv_proj")
        for key in ["gate_proj.weight", "up_proj.weight"]:
            if key in weight_key:
                return weight_key.replace(key, "gate_up_proj.weight")
        for key in ["q", "k", "v"]:
            if (
                "visual" in weight_key or "vision_tower" in weight_key
            ) and key in weight_key:
                return weight_key.replace(key, "qkv")
        return weight_key  # return full weight key

    def update_tensor_view(
        self,
        tensor_view: torch.Tensor,
        recv_tensor: torch.Tensor,
        inst_dest_name: str,
        **kwargs,
    ):
        tmp_recv_tensor = recv_tensor.to(tensor_view.dtype)
        if self.config.model_type == "gpt_oss" and "down_proj_bias" in inst_dest_name:
            assert (
                "parallel_dims" in kwargs
            ), "parallel_dims is required for update_tensor_view"
            tp_rank, _ = kwargs["parallel_dims"].tp_coord
            if tp_rank != 0:
                tmp_recv_tensor.zero_()
        tensor_view.copy_(tmp_recv_tensor)
